/*----------------------------------------------------------------------
  
 	OBJ2HEX

	Utility to convert an object file from MACRO-11 (PDP-11) to an
	Intel Hex file. All code must be in an ASECT.

	Usage: OBJ2HEX fname[.ext]
	  ext defaults to obj
	  Output file is fname.hex   

	Written by Mike Douglas
	
	 Ver     Date             Desc
	----   ----------    ------------------------
	 1.0   10/01/2024		Initial version

------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h> 
#include <string.h>
#include <ctype.h>

/* Constant definitions */

#define LINE_LEN 16			/* data bytes per hex output line */
#define FILENAME_LEN 100
#define GSD_BLOCK 1			/* ID for GSD (Global Symbol Directory) */
#define TXT_BLOCK 3			/* ID for TXT (code) data block */

/* Custom typedefs */

typedef uint8_t byte;		/* byte = unsigned 8 bit value */
typedef unsigned int uint;	/* uint = native unsigned integer */

/* Forward function references */

byte *write_hex_block (FILE *, byte *, uint);
byte *display_symbol_error (byte *, uint);


/* main */

int main (int argc, char *argv[])
{
	char	fileName[FILENAME_LEN];
	char	*extPtr;
	FILE	*inFile, *outFile;
	byte	*objPtr, *objEnd;
	uint	objLen, bytesRead, fbbLen, blockType;
	
/* Validate command line parameter */

	fprintf(stdout,"\nOBJ2HEX v1.0\n\n");
	if (argc != 2) {
		fprintf(stderr, "Usage: OBJ2HEX fname[.ext]\n\n"
							 "Default extension is \"obj\"\n"
							 "Output file is fname.hex\n");
		exit(1);
	}

/* Add .obj to file name if no extension provided. Use the same case as
      the user typed for Linux, etc. */

	strcpy(fileName, argv[1]);
	if ((extPtr = strrchr(fileName,'.')) == NULL) {
		extPtr = strrchr(fileName, '\0');			/* get pointer to end of fname */
		if (islower(*fileName))
			strcat(fileName, ".obj");
		else
			strcat(fileName, ".OBJ");
	}

/* Open the input file and put its length in objLen */

	if ((inFile = fopen(fileName, "rb")) == NULL) {
		fprintf(stderr,"Can't open file %s\n", fileName);
		exit(1);
	}

	fseek(inFile, 0, SEEK_END);
	objLen = ftell(inFile);

	if (objLen == -1) {
		fprintf(stderr,"File access failure %s\n", fileName);
		fclose(inFile);
		exit(1);
	}

/* Allocate memory based on file size and then read the entire object file
      into memory pointed to by objPtr */

	if ((objPtr = (byte *)malloc(objLen)) == NULL) {
		fprintf(stderr,"malloc failed\n");
		fclose(inFile);
		exit(1);
	}
	
	fseek(inFile, 0, SEEK_SET);					/* back to start of file */
	bytesRead = fread(objPtr, 1, objLen, inFile);
	fclose(inFile);
	if (bytesRead != objLen) {
		fprintf(stderr,"Error: File size is %u, but %u bytes read\n", objLen, bytesRead);
		exit(1);
	}
	objEnd = objPtr + objLen;						/* pointer to end of object file in RAM */

/* Create the output file name and open it */

	if (islower(*fileName))
		strcpy(extPtr, ".hex");						/* output file */
	else
		strcpy(extPtr, ".HEX");						/* output file */

	if ((outFile = fopen(fileName, "w")) == NULL) {
		fprintf(stderr,"Can't open the output file %s\n", fileName);
		exit(1);
	}

/* Scan the "Formatted Binary Blocks" (FBB) in objPtr looking for type 3
     data blocks (TXT) which contain the executable code and for type 1
	  blocks (GSD, Global Symbol Directory) which will contain symbols
	  only if they're undefined (error condition) */

	while (objPtr < objEnd) {
		if ((objPtr[0] == 1) && (objPtr[1] == 0)) {		/* FBB signature? */
			fbbLen = objPtr[2] + (objPtr[3] << 8);			/* get length of FBB */
			blockType = objPtr[4] + (objPtr[5] << 8);		/* get FBB type */

			if (blockType == TXT_BLOCK) 
				objPtr = write_hex_block(outFile, objPtr, fbbLen);

			else if (blockType == GSD_BLOCK)
				objPtr = display_symbol_error(objPtr, fbbLen);
		
			else								/* not a block type we use, skip it */
				objPtr += fbbLen + 1;
		}
		else 									/* FBB no found */
			objPtr++;						/* continue scan one byte forward */
	}

/* Finish output file and exit */

	fprintf(outFile, ":00000001FF\n");
	fclose(outFile);
	fprintf(stderr,"Hex file %s has been written\n\n", fileName);
	
	return 0;
}


/*----------------------------------------------------------------------
  Function write_hex_block
 		Writes the content of a TXT block to outFile in Intel hex format.
		Returns a pointer to the object array past the TXT block.
------------------------------------------------------------------------*/

byte *write_hex_block (FILE *outFile, byte *objPtr, uint fbbLen)
{
	uint	byteCount, dataAddr, checkSum, i;

	dataAddr = objPtr[6] + (objPtr[7] << 8);		/* address where data goes */
	objPtr += 8;											/* move to start of data */
	fbbLen -= 8;											/* data bytes left */

	while (fbbLen > 0) {									/* loop for all data byte in block */
		byteCount = fbbLen > LINE_LEN ? LINE_LEN : fbbLen;		/* output byte count */
		fbbLen -= byteCount;
		fprintf(outFile, ":%02X%04X00", byteCount, dataAddr);	/* write length and address */
		checkSum = byteCount + (dataAddr >> 8) + (dataAddr & 0xff);

		for (i=0; i<byteCount; i++) {					/* write the data bytes */
			fprintf(outFile,"%02X", *objPtr);
			checkSum += *objPtr++;
		}

		fprintf(outFile, "%02X\n", -checkSum & 0xff);	/* Intel hex checksum */
		dataAddr += byteCount;
	}
	return(objPtr);
}


/*----------------------------------------------------------------------
  Function display_symbol_error
		Scans a global symbol block (GSD) looking for any user program
		symbols. If found, these are an "undefined symbol" error the 
		user wants to know about. Returns a pointer to the object array
		past the GSD block.
------------------------------------------------------------------------*/

byte *display_symbol_error (byte *objPtr, uint fbbLen)
{

	static const char radix50[]=" ABCDEFGHIJKLMNOPQRSTUVWXYZ$.%0123456789";
	byte	*blockEnd;
	uint	symWord;								/* word contains 3 Radix-50 chars */
	uint	symCount=0;
	char	symStr[8];							/* ASCII string of the chars */


/* Loop through the 8 byte GSD entries looking for type 4 (global symbol)
		entries. If found, print as an undefined label. */

	blockEnd = objPtr + fbbLen;			/* pointer to the end of the block */
	objPtr += 6;								/* move to first symbol entry */

	while (objPtr < blockEnd) {
		if (objPtr[5] == 4){					/* "Global Symbol" type? */
			symWord = objPtr[0] + (objPtr[1] << 8);
			symStr[0] = radix50[symWord / (40*40)];
			symWord = symWord % (40*40);
			symStr[1] = radix50[symWord / 40];
			symStr[2] = radix50[symWord % 40];
	
			symWord = objPtr[2] + (objPtr[3] << 8);
			symStr[3] = radix50[symWord / (40*40)];
			symWord = symWord % (40*40);
			symStr[4] = radix50[symWord / 40];
			symStr[5] = radix50[symWord % 40];
			symStr[6] = '\0';

			fprintf(stderr,"Undefined Symbol: %s\n", symStr);
			symCount++;
		}
		objPtr += 8;
	}
	if (symCount != 0) 
		fprintf(stderr,"\n");

	return(objPtr);
}
